@@ -28,7 +28,7 @@ def with_summary(f):
     @functools.wraps(f)
     def wrapper(*args, **kwargs):
         summary = []
-        f(*args, summary=summary, **kwargs)
+        r = f(*args, summary=summary, **kwargs)
         if summary:
             summary = ['\nSummary of actions:'] + summary
             print "\n - ".join(summary)
@@ -36,4 +36,6 @@ def with_summary(f):
         else:
             print "No summary provided."
 
+        return r
+
     return wrapper
